from __future__ import print_function

from functools import partial

from collections import namedtuple

from twisted.internet.defer import Deferred
from twisted.internet import reactor

from types import ClassType

from .timeout import resolveTimeout

Result = namedtuple('Result', ['protocol', 'args', 'kw'])


class BoundHook(object):

    def __init__(self, hooked, original, instance):
        self.hooked = hooked
        self.original = original
        self.instance = instance

    def __call__(self, *args, **kw):
        result = self.original(self.instance, *args, **kw)
        self.hooked._called.callback(Result(self.instance, args, kw))
        return result

    def __getattr__(self, item):
        return getattr(self.hooked, item)


class Call(object):

    def __init__(self, *args, **kw):
        self.args = args
        self.kw = kw


class HookedCall(object):

    def __init__(self, class_, hook, decoder=Call):
        self._called = Deferred()
        self.original = getattr(class_, hook)
        self.decode = decoder

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return BoundHook(self, self.original, instance)

    def _reset(self, handler, timeout, result):
        if isinstance(result, Result):
            return handler(result)
        else:
            self._called = Deferred()
            self._called.addTimeout(timeout, reactor).addCallback(handler)
            return self._called

    def _expectCallback(self, handler, timeout):
        timeout = resolveTimeout(timeout)
        return self._called.addTimeout(timeout, reactor).addCallback(
            partial(self._reset, handler, timeout)
        )

    def called(self, decode=None, timeout=None):
        return self._expectCallback(
            lambda result: (decode or self.decode)(*result.args, **result.kw),
            timeout,
        )

    def protocol(self, timeout=None):
        return self._expectCallback(lambda result: result.protocol, timeout)


def hook(class_, *names):
    methods = {'__carly_hooked__': True}
    for name in names:
        d = Deferred()
        methods[name] = HookedCall(class_, name)

    if issubclass(class_, object):
        type_ = type
    else:  # pragma: no cover
        # some protocols don't have object has a base class!
        type_ = ClassType

    return type_('Hooked'+class_.__name__, (class_,), methods)


def hookMethod(class_, name, decoder=Call):
    if not getattr(class_, '__carly_hooked__', False):
        raise AssertionError("Can't hook a method on an unhooked class")
    setattr(class_, name, HookedCall(class_, name, decoder))



from functools import partial

from collections import namedtuple

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, gatherResults, maybeDeferred
from twisted.internet.protocol import Factory, ClientFactory

from .hook import hook
from .timeout import resolveTimeout

TCPServer = namedtuple('TCPServer', ['protocolClass', 'port'])
TCPClient = namedtuple('TCPClient', ['protocolClass', 'connection', 'protocol'])


class Context(object):

    def __init__(self):
        self.cleanups = {
            'connections': [],
            'listens': [],
        }

    def _cleanup(self, cleanups, timeout):
        deferreds = []
        for p in cleanups:
            d = p()
            deferreds.append(d)
            d.addTimeout(timeout, reactor)
        return gatherResults(deferreds)

    @inlineCallbacks
    def cleanup(self, timeout=None):
        timeout = resolveTimeout(timeout)
        yield self._cleanup(self.cleanups['connections'], timeout)
        yield self._cleanup(self.cleanups['listens'], timeout)

    def makeTCPServer(self, protocol, factory=None):
        protocolClass = hook(protocol, 'connectionLost')
        if factory is None:
            factory = Factory()
        factory.protocol = protocolClass
        port = reactor.listenTCP(0, factory)
        server = TCPServer(protocolClass, port)
        self.cleanupTCPServer(server)
        return server

    def cleanupTCPServer(self, server, timeout=None):
        timeout = resolveTimeout(timeout)
        self.cleanups['connections'].append(
            partial(server.protocolClass.connectionLost.called, timeout=timeout)
        )
        self.cleanups['listens'].append(
            partial(maybeDeferred, server.port.stopListening)
        )

    def makeTCPClient(self, protocol, port, factory=None, when='connectionMade'):
        protocolClass = hook(protocol, when, 'connectionLost')
        if factory is None:
            factory = ClientFactory()
        factory.protocol = protocolClass
        connection = reactor.connectTCP('localhost', port.getHost().port, factory)
        client = TCPClient(
            protocolClass, connection, getattr(protocolClass, when).protocol()
        )
        self.cleanupTCPClient(client)
        return client

    def cleanupTCPClient(self, client, timeout=None):
        timeout = resolveTimeout(timeout)
        self.cleanups['connections'].extend((
            partial(maybeDeferred, client.connection.disconnect),
            partial(client.protocolClass.connectionLost.called, timeout=timeout),
        ))





# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['carly']

package_data = \
{'': ['*']}

install_requires = \
['Twisted>=18.9,<19.0']

setup_kwargs = {
    'name': 'carly',
    'version': '0.2.1',
    'description': 'A tool for putting messages into and collecting responses from Twisted servers using real networking',
    'long_description': None,
    'author': 'Chris Withers',
    'author_email': 'chris@withers.org',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)
